/**
 * The main server entrypoint for Contaminate
 */

extern crate config;
extern crate pretty_env_logger;
extern crate serde;
extern crate serde_json;
extern crate surf;
extern crate tide;

use async_std::task;
use crypto::digest::Digest;
use crypto::sha2::Sha256;
use log::*;
use tide::{Request, Response};

use std::fs::File;
use std::io::Read;
use std::path::Path;

mod models;

/**
 * Load the settings based on the hierarchy.
 *
 * First we load the configuration file `contaminate.yml` if it exists
 * Then we look at environment variables.
 */
fn load_settings() -> config::Config {
    let mut settings = config::Config::default();
    settings.set_default("registry", "https://registry-1.docker.io")
        .expect("Could not set the default for `registry`");
    settings.set_default("layers_dir", "./layers.d")
        .expect("Could not set the default for `layers_dir`");

    settings
        .merge(config::File::with_name("contaminate").required(false))
        .expect("Failed to load settings in contaminate.ymll")
        .merge(config::Environment::with_prefix("CT"))
        .expect("Failed to load settings defined by CT_* env vars");

    debug!("Loaded configuration: {:?}", settings);

    return settings;
}

/**
 * AppState is a simple struct to carry information into request handlers
 */
#[derive(Clone, Debug)]
struct AppState {
    conf: config::Config,
}

impl AppState {
    /**
     * Helper f unction for fishing out the upstream registry URL
     */
    fn upstream(&self) -> String {
        self.conf.get_str("registry")
            .expect("`registry` not properly configured, must be a string")
    }


    /**
     * This function returns a true if the configured `layers_dir` has an override
     * for the given triplet of org/image:digest
     *
     * For example, if we have a `<layers_dir>/library/alpine/latest/` directory
     * with `*.tar.gz` files within it, then the function would return true.
     */
    fn override_exists(&self, org: &str, image: &str, digest: &str) -> bool {
        let layers_dir = self.conf.get_str("layers_dir")
            .expect("Unable to access `layers_dir` conf variable");

        info!("Looking in directory: {}", layers_dir);
        let path = Path::new(&layers_dir).join(org).join(image).join(digest);

        path.exists()
    }
}

/**
 * Make an authenticated GET request to the given registry
 */
async fn make_get_request(url: &str, authorization: &str, accepts: &str) -> Result<surf::Response, surf::Error> {
    debug!("Making GET request to {}", url);

    surf::get(url)
        .set_header("Authorization", authorization)
        .set_header("Accepts", accepts).await
}

/**
 * Simple convenience function for grabbing the HeaderValue out of the tide::Request and accessing
 * it as a string so contaminate can pass that to surf
 */
fn header_as_str(req: &Request<AppState>, header: &str, default: &str) -> String {
    req.header(header).map_or(default, |hv| hv.as_str()).to_string()
}

/**
 * Proxy the given response to the upstream registry and return the response
 * back to the client request it.
 */
async fn proxy_upstream(req: Request<AppState>) -> Result<Response, tide::Error> {
    let full_url = format!("{}{}", req.state().upstream(), req.url().path());
    info!("Proxying request upstream to {}", full_url);
     /*
      * We need to send the Authorization header along as well, otherwise
      * the upstream repository might complain that we're not authorized
      */
    let token = header_as_str(&req, "Authorization", "");
    let accepts = header_as_str(&req, "Accept", "");

    if let Ok(mut u_res) = make_get_request(&full_url, &token, &accepts).await {
        let body = u_res.body_string().await;
        debug!("Body received from {}", full_url);
        match body {
            Ok(body) => {
                trace!("upstream response for {}:\n{}", req.url(), body);
                /*
                 * If we don't explicitly set the content type here, the client will think
                 * that we're sending back a v1 manifest schema and complain about a "missing
                 * signature key"
                 */
                let content_type = u_res.header("Content-Type").map_or("text/plain", |h| h.as_str());
                let mut res = Response::new(u_res.status());
                res.set_body(body);
                res.insert_header("Content-Type", content_type);
                Ok(res)
            },
            Err(err) => {
                error!("Failed to make upstream request: {:?}", err);
                Ok(Response::new(500))
            },
        }
    }
    else {
        error!("Failed to make request upstream to {}", req.url());
        Ok(Response::new(500))
    }

}

/**
 * This function will fetch and manipulate the upstream manifest, typically
 * located at `/v2/myorg/myimage/manifests/latest`
 *
 * This will return a Response to the client which conforms to the manifest
 * specification.
 *
 * The manifests API is defined in the "Pulling an Image" part of the HTTP
 * registry API documentation, see here:
 *    <https://docs.docker.com/registry/spec/api/#pulling-an-image>
 *
 * This route will:
 *    - fetch the upstream manifest in order to fetch the image configuration
 *    - fetch the image configuration, which must have its diff_ids overwritten
 *    - re-compute the digest of the image configuration
 *    - grab all the layers which must be inserted into the manifest
 *    - generate the updated manifest
 */
async fn fetch_digest(req: Request<AppState>) -> Result<Response, tide::Error> {
    let org: String = req.param("org").unwrap_or("".to_string());
    let image: String = req.param("image").unwrap_or("".to_string());
    let digest: String = req.param("digest").unwrap_or("".to_string());

    if req.state().override_exists(&org, &image, &digest) {
        error!("Override exists, we should serve this request");
        let full_url = format!("{}{}", req.state().upstream(), req.url().path());
        let token = header_as_str(&req, "Authorization", "");
        let accepts = header_as_str(&req, "Accept", "");

        if let Ok(mut up_req) = make_get_request(&full_url, &token, &accepts).await {
            let manifest: models::Manifest = up_req.body_json().await.expect("Failed to parse response");
            debug!("Manifest: {:?}", manifest);
            Ok(Response::new(200))
        }
        else {
            error!("Failed to make an upstream request for {}", full_url);
            Ok(Response::new(500))
        }
    }
    else {
        error!("Override does not exist for {}/{}/{}, we should proxy upstream", org, image, digest);
        return proxy_upstream(req).await;
    }
}

async fn fetch_blob(req: Request<AppState>) -> Result<Response, tide::Error> {
    info!("fetch_blob: {}", req.url());
    Ok(Response::new(200))
}


/**
 * Perform the necessary SHA-256 checksum on the given Path
 */
fn checksum_for(path: &Path) -> Result<String, std::io::Error> {
    let mut file = File::open(path)?;
    let mut hasher = Sha256::new();
    let mut contents = String::new();

    file.read_to_string(&mut contents)?;

    hasher.input_str(&contents);
    Ok(hasher.result_str())
}

fn main() -> Result<(), std::io::Error> {
    pretty_env_logger::init();
    let conf = load_settings();
    let upstream_url = conf.get_str("registry")
        .expect("`registry` not properly configured, must be a string");

    info!("Starting with the following upstream: {}", upstream_url);

    let layers_dir = conf.get_str("layers_dir")
        .expect("`layers_dir` not properly configured, must be a string");
    let layers_dir = Path::new(&layers_dir);

    if ! layers_dir.is_dir() {
        error!("The `layers_dir` ({}) does not appear to be a directory", layers_dir.display());
        panic!("`layers_dir` must be a directory");
    }

    let state = AppState {
        conf: conf,
    };

    task::block_on(async {
        let mut app = tide::with_state(state);
        app.at("/").get(|_| async move { Ok("Hello, world!") });
        /*
         * This route works for "normal" images, which have name of org/image
         */
        app.at("/v2/:org/:image/manifests/:digest").get(fetch_digest);
        /*
         * This route handles images which look like "official" images,
         * such as `alpine:latest`, which _actually_ maps to `library/alpine:latest`
         * in DockerHub
         */
        //app.at("/v2/:image/manifests/:digest").get(fetch_digest);

        app.at("/v2/:org/:image/blobs/:sha").get(fetch_blob);
        /*
         * The catch-all for the remainder of the v2 API calls should proxy to
         * the upstream repository, since Contaminate does not implement a full
         * registry API
         */
        app.at("/v2/*").get(proxy_upstream);
        app.listen("127.0.0.1:9090").await?;
        Ok(())
    })
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn simple_checksum() {
        let path = Path::new("./LICENSE.txt");
        let checksum = checksum_for(&path).expect("Failed to read");
        assert_eq!("0d96a4ff68ad6d4b6f1f30f713b18d5184912ba8dd389f86aa7710db079abcb0", checksum);
    }

    #[test]
    fn checksum_for_invalid_path() {
        let path = Path::new("./non-existent-womp-womp");
        let result = checksum_for(&path);
        assert!(result.is_err());
    }

    #[test]
    fn override_exists_false() {
        let conf = load_settings();
        let state = AppState { conf };
        assert_eq!(false, state.override_exists("jenkins", "jenkins", "latest"));
    }


    #[test]
    fn override_exists_true() {
        let conf = load_settings();
        let state = AppState { conf };
        assert!(state.override_exists("library", "alpine", "latest"));
    }
}
